/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "maya/maya.pmh"

#define FE_JCO_DEBUG		FALSE
#define FE_JCO_REFERENCE	FALSE

using namespace fe;

void JointCombineOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<String>("root")="joints";
	catalog<String>("root","label")="Root Node Name";

	//* not read; just for update dependency
	catalog< sp<Component> >("Input Trigger");

#if FE_JCO_REFERENCE
	catalog<bool>("reference")=true;

	//* TODO optional ref poly input
	catalog< sp<Component> >("Reference Joints");
	catalog<bool>("Reference Joints","optional")=true;
#endif

	//* no copy
	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";

	m_spScope=registry()->create("Scope");
}

void JointCombineOp::handle(Record& a_rSignal)
{
	const String rootName=catalog<String>("root");

#if FE_JCO_DEBUG
	feLog("JointCombineOp::handle\n");
#endif

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	m_spJointAccessible=registry()->create("*.SurfaceAccessibleJoint");
	if(m_spJointAccessible.isNull())
	{
		return;
	}

	m_spJointAccessible->bind(m_spScope);
	m_spJointAccessible->startJoints();

#if FE_JCO_REFERENCE
	sp<SurfaceAccessibleI> spRefAccessible;
	if(catalog<bool>("reference"))
	{
		access(spRefAccessible,"Reference Joints",e_quiet);
	}

	m_primitiveIndexMap.clear();

	if(spRefAccessible.isValid())
	{
		sp<SurfaceAccessorI> spInputRefName;

		if(!access(spInputRefName,spRefAccessible,e_primitive,"name") ||
				!access(m_spInputRefX,spRefAccessible,e_primitive,"refX") ||
				!access(m_spInputRefY,spRefAccessible,e_primitive,"refY") ||
				!access(m_spInputRefZ,spRefAccessible,e_primitive,"refZ") ||
				!access(m_spInputRefT,spRefAccessible,e_primitive,"refT"))
		{
			spRefAccessible=NULL;
		}
		else
		{
			const U32 primitiveCount=spInputRefName->count();
			for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;
					primitiveIndex++)
			{
				const String name=spInputRefName->string(primitiveIndex);
				m_primitiveIndexMap[name]=primitiveIndex;
			}
		}
	}

	if(spRefAccessible.isNull())
	{
		m_spInputRefX=NULL;
		m_spInputRefY=NULL;
		m_spInputRefZ=NULL;
		m_spInputRefT=NULL;
	}
#endif

	//* TODO is there an easier way to get a node object by name?
	MSelectionList originalList;
	MGlobal::getActiveSelectionList(originalList);

	MGlobal::selectByName(rootName.c_str(),MGlobal::kReplaceList);

	MSelectionList selectionList;
	MGlobal::getActiveSelectionList(selectionList);

	MGlobal::setActiveSelectionList(originalList);

	MStatus status;
	MObject rootObject;
	if(selectionList.length())
	{
		selectionList.getDependNode(0,rootObject);
	}
	if(rootObject.isNull())
	{
		String message;
		message.sPrintf("root node not found: %s",
				status.errorString().asChar());
		catalog<String>("error")+=message;
		return;
	}

	m_animWorldMap.clear();
	setIdentity(m_animWorldMap[""]);

	m_refWorldMap.clear();
	setIdentity(m_refWorldMap[""]);

	scanDag(rootObject,"");

	spOutputAccessible->copy(m_spJointAccessible);

#if FE_JCO_DEBUG
	feLog("JointCombineOp::handle done\n");
#endif
}

void JointCombineOp::scanDag(MObject a_parent,String a_grandparent)
{
	const Color yellow(1.0,1.0,0.0);

	if(a_parent.isNull())
	{
		return;
	}

	MFnDagNode dagParent(a_parent);
	const String name=dagParent.name().asChar();

	SpatialTransform xformLocal;

	MStatus status;
	MFnIkJoint fnJoint(a_parent,&status);
	if(status)
	{
		const MVector trans=fnJoint.getTranslation(MSpace::kTransform);

		MQuaternion rotation;
		fnJoint.getRotation(rotation,MSpace::kTransform);

		MQuaternion orientation;
		status=fnJoint.getOrientation(orientation);

		const MQuaternion rotateOrient=
				fnJoint.rotateOrientation(MSpace::kTransform);

		const SpatialVector relativeTranslation(trans[0],trans[1],trans[2]);
		const SpatialQuaternion relativeRotation(
				rotation[0],rotation[1],rotation[2],rotation[3]);
		const SpatialQuaternion relativeRotateOrient(
				rotateOrient[0],rotateOrient[1],
				rotateOrient[2],rotateOrient[3]);
		const SpatialQuaternion relativeOrientation(
				orientation[0],orientation[1],
				orientation[2],orientation[3]);

		SpatialTransform relativeMove;
		setIdentity(relativeMove);
		translate(relativeMove,relativeTranslation);

		xformLocal=SpatialTransform(relativeRotation)*relativeMove;

#if FE_JCO_DEBUG
		if(name=="l_uparm")
		{
			const SpatialEuler euler=SpatialTransform(relativeRotation);

			feLog("for \"%s\" t %s r %s e %s\n",
					dagParent.name().asChar(),
					c_print(relativeTranslation),
					c_print(relativeRotation));
			feLog("  e %.6G %.6G %.6G ro %s o %s\n%s\n",
					euler[0]/degToRad,
					euler[1]/degToRad,
					euler[2]/degToRad,
					c_print(relativeRotateOrient),
					c_print(relativeOrientation),
					c_print(xformLocal));
		}
#endif
	}
	else
	{
		setIdentity(xformLocal);
	}

	const SpatialTransform animWorld=xformLocal*m_animWorldMap[a_grandparent];

	m_animWorldMap[name]=animWorld;

#if FE_JCO_REFERENCE

	const U32 primitiveIndex=m_primitiveIndexMap[name];

	SpatialTransform xformLocalRef;
	if(m_spInputRefX.isValid())
	{
		xformLocalRef.column(0)=m_spInputRefX->spatialVector(primitiveIndex);
		xformLocalRef.column(1)=m_spInputRefY->spatialVector(primitiveIndex);
		xformLocalRef.column(2)=m_spInputRefZ->spatialVector(primitiveIndex);
		xformLocalRef.column(3)=m_spInputRefT->spatialVector(primitiveIndex);
	}
	else
	{
		xformLocalRef=xformLocal;
	}

//	feLog("xformLocalRef\n%s\n",c_print(xformLocalRef));
//	feLog("xformLocal\n%s\n",c_print(xformLocal));

	const SpatialTransform refWorld=xformLocalRef*m_refWorldMap[a_grandparent];

	m_refWorldMap[name]=refWorld;

	if(!a_grandparent.empty())
	{
		m_spJointAccessible->addJoint(name,a_grandparent,
				xformLocal,animWorld,refWorld,yellow);
	}
#else
	if(!a_grandparent.empty())
	{
		m_spJointAccessible->addJoint(name,a_grandparent,
				xformLocal,animWorld,yellow);
	}
#endif

	const U32 childCount=dagParent.childCount(&status);
	for(U32 childIndex=0;childIndex<childCount;childIndex++)
	{
		MObject child=dagParent.child(childIndex,&status);
		MFnDagNode dagChild(child);
//		feLog("parent \"%s\" child \"%s\"\n",
//				dagParent.name().asChar(),dagChild.name().asChar());

		scanDag(child,name);
	}
}
