/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "operator/operator.pmh"

namespace fe
{

void RasterOp::initialize(void)
{
	catalog<String>("icon")="FE_beta";

	catalog<String>("valueAttr")="value";
	catalog<String>("valueAttr","label")="Value Attribute";
	catalog<String>("valueAttr","hint")=
			"Name of point attribute to place results.";

	catalog<String>("pattern")="PART";
	catalog<String>("pattern","label")="Replace Pattern";
	catalog<String>("pattern","hint")=
			"Replace this string in the following filename with"
			" each unique partition name.";

	catalog<bool>("colorize")=false;
	catalog<String>("colorize","label")="Colorize";

	catalog<bool>("save")=false;
	catalog<bool>("save","joined")=true;
	catalog<String>("save","label")="Save";

	catalog<String>("savefile")="raster.PART.$F.png";
	catalog<String>("savefile","label")="File";

	catalog< sp<Component> >("Input Surface");
}

void RasterOp::handle(Record& a_rSignal)
{
	const String valueAttr=catalog<String>("valueAttr");
	const String pattern=catalog<String>("pattern");
	const BWORD colorize=catalog<bool>("colorize");
	const BWORD save=catalog<bool>("save");
	const String savefile=save? catalog<String>("savefile"): "";

	if(save && savefile.empty())
	{
		catalog<String>("warning")+="no filename for saving image;";
		return;
	}

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessorI> spOutputColor;
	if(!access(spOutputColor,spOutputAccessible,
			e_point,"Cd")) return;

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Surface")) return;

	sp<SurfaceAccessorI> spInputWidth;
	if(!access(spInputWidth,spInputAccessible,
			e_detail,"raster_width")) return;

	sp<SurfaceAccessorI> spInputHeight;
	if(!access(spInputHeight,spInputAccessible,
			e_detail,"raster_height")) return;

	sp<SurfaceAccessorI> spInputMin;
	if(!access(spInputMin,spInputAccessible,
			e_detail,"raster_min")) return;

	sp<SurfaceAccessorI> spInputMax;
	if(!access(spInputMax,spInputAccessible,
			e_detail,"raster_max")) return;

	sp<SurfaceAccessorI> spInputPartAttr;
	if(!access(spInputPartAttr,spInputAccessible,
			e_detail,"raster_partition")) return;

	sp<SurfaceAccessorI> spInputValue;
	if(!access(spInputValue,spInputAccessible,e_point,valueAttr)) return;

	const I32 width=spInputWidth->integer(0);
	const I32 height=spInputHeight->integer(0);
	const Real min=spInputMin->real(0);
	const Real max=spInputMax->real(0);
	const String partitionAttr=spInputPartAttr->string(0);

	sp<SurfaceAccessorI> spInputPartition;
	if(!partitionAttr.empty())
	{
		if(!access(spInputPartition,spInputAccessible,
				e_point,partitionAttr)) return;
	}

	sp<ImageI> spImage;
	if(save)
	{
		String format;

		String buffer=savefile;
		String token;
		while(!(token=buffer.parse("",".")).empty())
		{
			format=token;
		}

		if(!format.empty())
		{
			spImage=registry()->create("ImageI.*.*."+format);
		}
		if(spImage.isNull())
		{
			spImage=registry()->create("ImageI");
		}
		if(spImage.isNull())
		{
			catalog<String>("warning")+="could not initialize output image;";
		}

		feLog("RasterOp::handle save using \"%s\"\n",spImage->name().c_str());
	}

	const U32 pointCount=spInputValue->count();

	//* TODO could use std::unordered_set when available

	std::set<String> partitionSet;
	if(spInputPartition.isValid())
	{
		for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			const String partition=spInputPartition->string(pointIndex);
			if(partitionSet.find(partition)==partitionSet.end())
			{
				partitionSet.insert(partition);
			}
		}
	}
	else
	{
		partitionSet.insert("");
	}

	U8 data[height][width][3];

	//* TODO create each image in parallel

	for(std::set<String>::iterator it=partitionSet.begin();
			it!=partitionSet.end();it++)
	{
		const String partition= *it;
		const String partitionFile=savefile.replace(pattern,partition);

		feLog("RasterOp::handle partition \"%s\" file \"%s\"\n",
				partition.c_str(),partitionFile.c_str());

		U32 pixelIndex=0;

		for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			if(spInputPartition.isValid() &&
					spInputPartition->string(pointIndex)!=partition)
			{
				continue;
			}

			const U32 uIndex=pixelIndex%width;
			const U32 vIndex=pixelIndex/width;

			if(uIndex>=width || vIndex>=height)
			{
				feLog("%d u %d/%d v %d/%d\n",
						pointIndex,uIndex,width,vIndex,height);
				continue;
			}

			U8* pPixel=data[vIndex][uIndex];

			//* TODO verify correct uv
//			const Real uCoord=uIndex/(width-1.0);
//			const Real vCoord=vIndex/(height-1.0);

			const Real value=spInputValue->real(pointIndex);

			I32 byte=255*(value-min)/(max-min);
			if(byte<0)
			{
				byte=0;
			}
			if(byte>255)
			{
				byte=255;
			}

			pPixel[0]=byte;

			if(colorize)
			{
				pPixel[1]=value>0.0? 0: byte;
				pPixel[2]=byte<5? 255: 0;
			}
			else
			{
				pPixel[1]=byte;
				pPixel[2]=byte;
			}

			const Color color=(1.0/255.0)*Color(pPixel[0],pPixel[1],pPixel[2]);
			spOutputColor->set(pointIndex,color);

//			feLog("%d u %d/%d v %d/%d bytes %d %d %d\n",
//					pointIndex,uIndex,width,vIndex,height,
//					pPixel[0],pPixel[1],pPixel[2]);

			pixelIndex++;
		}

		if(spImage.isValid())
		{
			spImage->createSelect();
			spImage->setFormat(ImageI::e_rgb);
			spImage->resize(width,height,1);
			spImage->replaceRegion(0,0,0,width,height,1,data);

			if(!spImage->save(partitionFile))
			{
				catalog<String>("warning")+=
						"failed to save \"" + partitionFile +"\";";
			}

			//* TODO try without clearing last; need ref count on ImageRaw data
			spImage->unload(0);
		}
	}
}

} // namespace
